# These have to be synced with the stdlib.pxi
import asyncio
import collections
import concurrent.futures
import functools
import itertools
import os
import signal
import socket
import subprocess
import sys
import threading
